# Copyright (c) 2024 Bytedance Ltd. and/or its affiliates

#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at

#      http://www.apache.org/licenses/LICENSE-2.0

#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import subprocess
from pathlib import Path

from git import GitError, Repo


def exists(path: Path) -> bool:
    if not path.exists():
        return False

    return path.is_dir() and path.joinpath(".git").exists()


def is_clean(path: Path) -> tuple[bool, str]:
    if not exists(path):
        return False, f"Repository not found: {path}"

    result = subprocess.run(
        ["git", "-C", str(path), "status", "--porcelain"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )

    if result.returncode != 0:
        return False, f"Git error occurred: {result.stderr.decode()}"

    if result.stdout.decode().strip() == "":
        return True, ""

    return False, f"Repository is not clean: {path}"


def clone_repository(destination: Path, org: str, repo: str):
    repo_url = f"https://github.com/{org}/{repo}.git"
    repo_path = destination / repo
    subprocess.run(
        ["git", "clone", repo_url, str(repo_path)],
        check=True,
    )


def get_all_commit_hashes(repo_path: Path, logger: logging.Logger) -> set[str]:
    try:
        repo = Repo(repo_path)

        commit_hashes = set()
        for commit in repo.iter_commits("--all"):
            commit_hashes.add(commit.hexsha)

        return commit_hashes
    except GitError as e:
        logger.error(f"Git error occurred: {e}")
        return set()

def clean(repo_dir):
    subprocess.run(["git", "reset", "--hard"], cwd=repo_dir, check=True)
    subprocess.run(["git", "clean", "-fd"], cwd=repo_dir, check=True)
